

#include <Python.h>
#include "ann-cc.h"

#define PyANN_AsData(x) (((PyANNObject*)x)->_p)
#define RETURN_PY_NONE	return (Py_INCREF(Py_None), Py_None)

static PyTypeObject PyANN_Type;


typedef struct {
  PyObject_HEAD
  struct ann_data_t* _p;
} PyANNObject;


PyObject*
ann_new_ann(PyObject* self, PyObject* args)
{
  PyANNObject* ann;
  int alloc_num_pts;
  int dims;

  if (!PyArg_ParseTuple(args, "ii:ann_new_ann", &alloc_num_pts, &dims))
    return NULL;

  ann = PyObject_New(PyANNObject, &PyANN_Type);

  assert(ann);
  ann->_p = ann_cc_make_ann(alloc_num_pts, dims);

  // if the call failed, let Python know
  if (!ann->_p) {
    return NULL;
  }
  
  return (PyObject*)ann;
}


void
ann_ann_dealloc(PyObject* self)
{
  assert(self);
  ann_cc_del_ann(PyANN_AsData(self));
  PyObject_Del(self);
}


static PyObject*
ann_add_pt(PyObject* self, PyObject* args)
{
  int len;
  double arr[100];
  PyObject* seq;
  unsigned i;
  
  assert(PyTuple_Size(args) == 1);
  seq = PyTuple_GetItem(args, 0);
  len = PySequence_Size(seq);

  if (len < 0) {
    fprintf(stderr, "ann_add_pt(): unable to get the length of the sequence\n");
    return NULL;
  }

  if (len > 100) {
    fprintf(stderr, "ann_add_pt(): sequence too long; ");
    fprintf(stderr, "implementation can only handle length 100\n");
    return NULL;
  }

  for (i=0; i<len; i++) {
    PyObject* v = PySequence_GetItem(seq, i);
    arr[i] = PyFloat_AsDouble(v);
    Py_XDECREF(v);
  }

  ann_cc_add_pt(PyANN_AsData(self), arr, len);

  assert(!PyErr_Occurred());

  RETURN_PY_NONE;
}


static PyObject*
ann_search(PyObject* self, PyObject* args)
{
  unsigned nargs;
  double err_bound = 0;
  unsigned len;
  double arr[100];
  int num_res;
  PyObject* seq;
  unsigned i;

  nargs = PyTuple_Size(args);
  assert(nargs == 2 || nargs == 3); /* TODO: softer recovery */
  seq = PyTuple_GetItem(args, 0);
  num_res = PyInt_AsLong(PyTuple_GetItem(args, 1));
  if (nargs == 3)
    err_bound = PyFloat_AsDouble(PyTuple_GetItem(args, 2));

  if (num_res <= 0) {
    fprintf(stderr, "ann_search(): `num_res' should be positive!\n");
    return NULL;
  }
    
  len = PySequence_Size(seq);

  if (len < 0) {
    fprintf(stderr, "ann_search(): unable to get the length of the sequence\n");
    return NULL;
  }

  if (len > 100) {
    fprintf(stderr, "ann_search(): sequence too long; ");
    fprintf(stderr, "implementation can only handle length 100\n");
    return NULL;
  }

  for (i=0; i<len; i++) {
    PyObject* v = PySequence_GetItem(seq, i);
    arr[i] = PyFloat_AsDouble(v);
    Py_XDECREF(v);
  }

  num_res = ann_cc_search(PyANN_AsData(self), arr, len, num_res, err_bound);

  assert(!PyErr_Occurred());

  return Py_BuildValue("i", num_res);
}


static PyObject*
ann_get_nnidx(PyObject* self, PyObject* args)
{
  int idx;
  int nnidx;
  
  if(!PyArg_ParseTuple(args, "i", &idx))
    return NULL;

  nnidx = ann_cc_get_nnidx(PyANN_AsData(self), idx);
  
  return Py_BuildValue("i", nnidx);
}


static PyObject*
ann_get_dist(PyObject* self, PyObject* args)
{
  int idx;
  double dist;
  
  if(!PyArg_ParseTuple(args, "i", &idx))
    return NULL;

  dist = ann_cc_get_dist(PyANN_AsData(self), idx);
  
  return Py_BuildValue("d", dist);
}


static PyMethodDef annobj_methods[] = {
  { "add_pt", ann_add_pt, METH_VARARGS },
  { "search", ann_search, METH_VARARGS },
  { "get_nnidx", ann_get_nnidx, METH_VARARGS },
  { "get_dist", ann_get_dist, METH_VARARGS },
  { NULL, NULL }  
};


static PyObject*
ann_getattr(PyObject* self, char* attrname)
{
  return Py_FindMethod(annobj_methods, self, attrname);
}


static PyTypeObject PyANN_Type = {
  PyObject_HEAD_INIT(NULL)
  0,
  "Ann",
  sizeof(PyANNObject),
  0,
  ann_ann_dealloc,
  0,
  ann_getattr,
};


static PyMethodDef annmod_methods[] = {
  { "ANN", ann_new_ann, METH_VARARGS },
  { NULL, NULL }
};


DL_EXPORT(void)
initann(void)
{
  PyANN_Type.ob_type = &PyType_Type;
  Py_InitModule("ann", annmod_methods);
}

